#include <algorithm>
#include <cassert>
#include <fstream>
#include <iostream>
#include <vector>
#include <set>

using namespace std;

const char DELIM = 0;
const string DELIMS = "\0\0\0\0";

string read_file(const char *filename) {
    ifstream in(filename, ios::in | ios::binary);
    string contents;
    in.seekg(0, ios::end);
    contents.resize(in.tellg());
    in.seekg(0, ios::beg);
    in.read(&contents[0], contents.size());
    in.close();
    return contents;
}

struct suffix {
    int index;
    pair<uint32_t , uint32_t> rank;
};

string bwt(const string& text) {
    int N = text.size() + 1;
    string txt = text + DELIMS;
    vector<suffix> suffixes(N);

    for (int i = 0; i < N; i++) {
        suffixes[i].index = i;
        suffixes[i].rank.first = txt[i];
        suffixes[i].rank.second = txt[i + 1];
    }

    sort(suffixes.begin(), suffixes.end(), [](suffix &a, suffix &b) { return a.rank < b.rank; });

    vector<int> indices(N);
    for (int k = 2; k < N; k = k * 2) {
        uint32_t rank = 0;
        uint32_t prev_rank = suffixes[0].rank.first;
        suffixes[0].rank.first = rank;
        indices[suffixes[0].index] = 0;

        for (int i = 1; i < N; i++) {
            if (suffixes[i].rank.first == prev_rank &&
                suffixes[i].rank.second == suffixes[i - 1].rank.second) {
                suffixes[i].rank.first = rank;
            } else {
                prev_rank = suffixes[i].rank.first;
                suffixes[i].rank.first = ++rank;
            }

            indices[suffixes[i].index] = i;
        }
        
        for (uint32_t i = 0; i < N; i++) {
            uint32_t next_index = suffixes[i].index + k;
            suffixes[i].rank.second = (next_index < N) ? suffixes[indices[next_index]].rank.first : -1;
        }

        sort(suffixes.begin(), suffixes.end(), [](suffix &a, suffix &b) { return a.rank < b.rank; });
    }

    string bwt;
    bwt.resize(N);

    for (uint32_t i = 0; i < N; i++) {
        int32_t end_index = suffixes[i].index - 1;
        if (end_index < 0) end_index += N;
        bwt[i] = txt[end_index];
    }

    return bwt;
}

string ibwt(const string& bwt) {
    int N = bwt.size();
    vector<size_t> F(256), rank;
    rank.reserve(N);
    for (char x : bwt) rank.push_back(F[x]++);
    int start = 0;
    for (int c = 0; c < 256; c++) {
        int tmp = F[c];
        F[c] = start;
        start += tmp;
    }

    string result;
    result.resize(--N);
    int i = 0;
    char c = bwt[0];
    while (N) {
        result[--N] = c;
        i = F[c] + rank[i];
        c = bwt[i];
    }
    return result;
}

int count_runs(string &bwt) {
    int prev = -1, count = 0;
    for (const char c : bwt) {
        if (prev != c) count++;
        prev = c;
    }
    return count;
}

int main(int argc, char** argv) {
    const auto text = read_file("chr21.txt");
    //string text = "IF A DOG CHEWS SHOES WHOSE SHOES DOES HE CHOOSE";
    
    string encoded = bwt(text);
    string decoded = ibwt(encoded);
    
    assert(decoded == text);

    cout << "#runs: " << count_runs(encoded) << '\n';
    return 0;
}
