#include <algorithm>
#include <cassert>
#include <fstream>
#include <iostream>
#include <vector>
#include <set>
#include <string_view>

using namespace std;

const char DELIM = 0;
const string DELIMS = "\0\0\0\0\0\0\0\0";

string read_file(const char *filename) {
    ifstream in(filename, ios::in | ios::binary);
    string contents;
    in.seekg(0, ios::end);
    contents.resize(in.tellg());
    in.seekg(0, ios::beg);
    in.read(&contents[0], contents.size());
    in.close();
    return contents;
}

string bwt(string text) {
    int n = text.size();
    string txt = text + DELIMS;
    
    std::vector<uint32_t> result;
    result.resize(n);
    std::vector<string_view> suffixes(n);

    for (int i = 0; i < n; i++) {
       suffixes[i] = string_view(&txt[i], n-i);
    }
    sort(suffixes.begin(), suffixes.end());

    string bwt;
    bwt.resize(n);

    for (int i = 0; i < n; i++) {
        int32_t end_index = (n - suffixes[i].size())  - 1;
        if (end_index < 0) end_index += n;
        bwt[i] = txt[end_index];
    }
    return bwt;
}

string ibwt(const string& bwt) {
    int N = bwt.size();
    vector<size_t> F(256), rank;
    rank.reserve(N);
    for (char x : bwt) rank.push_back(F[x]++);
    int start = 0;
    for (int c = 0; c < 256; c++) {
        int tmp = F[c];
        F[c] = start;
        start += tmp;
    }

    string result;
    result.resize(--N);
    int i = 0;
    char c = bwt[0];
    while (N) {
        result[--N] = c;
        i = F[c] + rank[i];
        c = bwt[i];
    }
    return result;
}

int count_runs(string &bwt) {
    int prev = -1, count = 0;
    for (const char c : bwt) {
        if (prev != c) count++;
        prev = c;
    }
    return count;
}

int main(int argc, char** argv) {
    const auto text = read_file("chr21.txt");
    //string text = "IF A DOG CHEWS SHOES WHOSE SHOES DOES HE CHOOSE";
    
    string encoded = bwt(text+DELIM);
    string decoded = ibwt(encoded);
    
    assert(decoded == text);

    cout << "#runs: " << count_runs(encoded) << '\n';
    return 0;
}
