#include <algorithm>
#include <cassert>
#include <fstream>
#include <iostream>
#include <vector>
#include <set>

using namespace std;

const char DELIM = 0;
const string DELIMS = "\0\0\0\0\0\0\0\0";
const int PRESORT_LEN = 256;

string read_file(const char *filename) {
    ifstream in(filename, ios::in | ios::binary);
    string contents;
    in.seekg(0, ios::end);
    contents.resize(in.tellg());
    in.seekg(0, ios::beg);
    in.read(&contents[0], contents.size());
    in.close();
    return contents;
}

struct suffix {
    uint64_t index;
    uint64_t rank;
};

void print_suffixes(vector<suffix> &suffixes, string &txt, vector<int> &indices, int k) {
  int N = suffixes.size();
  for (int i=0; i<suffixes.size(); i++) {
      int next_index = suffixes[i].index+k;
      printf("%3ld %2d %2ld - %2ld %2ld %s\n",
          suffixes[i].index,
          next_index,
          (next_index < N) ? suffixes[indices[next_index]].rank >> 32 : 0,
          (suffixes[i].rank >> 32), (suffixes[i].rank & 0xFFFFFFFF), 
          string_view(&txt[suffixes[i].index], min(N-i, 10)).data());
  }
}

void print_buckets(vector<uint64_t> &buckets) {
    cout << "\nbuckets:\n";
    for (int i=0; i<buckets.size(); i+=2) {
      cout << "[" << buckets[i] << ", " << buckets[i+1] << "] ";
    }
    cout << "\n";
}

string bwt(const string& text) {
    int N = text.size() + 1;
    string txt = text + DELIMS;

    std::vector<string_view> pre(N);
    for (int i = 0; i < N; i++) {
       pre[i] = string_view(&txt[i], min(N-i, PRESORT_LEN));
    }
    sort(pre.begin(), pre.end());

    vector<suffix> suffixes(N);
    vector<int> indices(N);
    vector<uint64_t> buckets; buckets.reserve(N);
    vector<uint64_t> buckets2; buckets2.reserve(N);
    bool bucket_started = false;
    int start = 0;

    {
        int j = (N - pre[0].size());
        suffixes[0].index = j;
        suffixes[0].rank = 0;
    }
    uint64_t R = 0;
    for (uint64_t i = 1; i < N; i++) {
        int j = (&pre[i][0] - &txt[0]);
        suffixes[i].index = j;
        if (pre[i] == pre[i-1]) {
            if (!bucket_started) {
                bucket_started = true;
                start = i - 1;
            }
            suffixes[i].rank = R << 32;
        } else {
            if (bucket_started) {
                bucket_started = false;
                buckets.push_back(start);
                buckets.push_back(i);
            }
            R = i;
            suffixes[i].rank = R << 32;
        }
        indices[suffixes[i].index] = i;
    }
    if (bucket_started) {
        buckets.push_back(start);
        buckets.push_back(N);
    }

    for (int j = 0; j < buckets.size(); j+=2) {
        for (int i = buckets[j]; i < buckets[j+1]; i++) {
            uint64_t next_index = suffixes[i].index + PRESORT_LEN;
            suffixes[i].rank |= (next_index < N) ? suffixes[indices[next_index]].rank >> 32 : 0;
        }
    }
    
    // print_suffixes(suffixes, txt, indices, PRESORT_LEN);
    // print_buckets(buckets);

    for (int k = PRESORT_LEN; !buckets.empty(); ) {
        buckets2.clear();

        for (int j = 0; j < buckets.size(); j+=2) {
            sort(&suffixes[buckets[j]], &suffixes[buckets[j+1]], [](suffix a, suffix b) { return a.rank < b.rank; });

            bool bucket_started = false;
            uint64_t start = 0;
            uint64_t R = buckets[j], prev_rank = suffixes[buckets[j]].rank;
            // suffixes[buckets[j]].rank &= 0xFFFFFFFF00000000; 

            indices[suffixes[buckets[j]].index] = buckets[j];
            for (uint64_t i = buckets[j]+1; i < buckets[j+1]; i++) {
                if (suffixes[i].rank == prev_rank) {
                    if (!bucket_started) {
                        bucket_started = true;
                        start = i - 1;
                    }
                    suffixes[i].rank = (suffixes[i].rank & 0xFFFFFFFF) | R << 32;
                } else {
                    if (bucket_started) {
                        bucket_started = false;
                        buckets2.push_back(start);
                        buckets2.push_back(i);
                    }
                    prev_rank = suffixes[i].rank;
                    R = i;
                    suffixes[i].rank = (suffixes[i].rank & 0xFFFFFFFF) | R << 32;
                }
                indices[suffixes[i].index] = i;
            }
            for (uint64_t i = buckets[j]; i < buckets[j+1]; i++) {
                if (suffixes[i].index >= k) {
                    uint64_t prev_index = suffixes[i].index - k;
                    // printf("%ld %2d %2d-> ", prev_index, suffixes[indices[prev_index]].rank >> 32, suffixes[indices[prev_index]].rank & 0xFFFF);
                    suffixes[indices[prev_index]].rank = (suffixes[indices[prev_index]].rank & 0xFFFFFFFF00000000) | (suffixes[i].rank >> 32);
                    // printf("%2d %2d\n", suffixes[indices[prev_index]].rank >> 32, suffixes[indices[prev_index]].rank & 0xFFFF);
                }                
            }
            if (bucket_started) {
                buckets2.push_back(start);
                buckets2.push_back(buckets[j+1]);
            }
        }
        
        swap(buckets, buckets2);

        k = k * 2;

        for (int j = 0; j < buckets.size(); j+=2) {
            for (int i = buckets[j]; i < buckets[j+1]; i++) {
                int next_index = suffixes[i].index + k;
                if (next_index < N) {
                    suffixes[i].rank = (suffixes[i].rank & 0xFFFFFFFF00000000 ) | (suffixes[indices[next_index]].rank >> 32);
                }
            }
        }

        // print_suffixes(suffixes, txt, indices, k);
        // print_buckets(buckets);

        // return txt;
    }

    string bwt;
    bwt.resize(N);

    for (int i = 0; i < N; i++) {
        int32_t end_index = suffixes[i].index - 1;
        if (end_index < 0) end_index += N;
        bwt[i] = txt[end_index];
    }

    return bwt;
}

string ibwt(const string& bwt) {
    int N = bwt.size();
    vector<size_t> F(256), rank;
    rank.reserve(N);
    for (char x : bwt) rank.push_back(F[x]++);
    int start = 0;
    for (int c = 0; c < 256; c++) {
        int tmp = F[c];
        F[c] = start;
        start += tmp;
    }

    string result;
    result.resize(--N);
    int i = 0;
    char c = bwt[0];
    while (N) {
        result[--N] = c;
        i = F[c] + rank[i];
        c = bwt[i];
    }
    return result;
}

int count_runs(string &bwt) {
    int prev = -1, count = 0;
    for (const char c : bwt) {
        if (prev != c) count++;
        prev = c;
    }
    return count;
}

int main(int argc, char** argv) {
    const auto text = read_file("chr21.txt");
    // string text = "IF A DOG CHEWS SHOES WHOSE SHOES DOES HE CHOOSE";
    
    string encoded = bwt(text);
    string decoded = ibwt(encoded);
    
    assert(decoded == text);

    cout << "#runs: " << count_runs(encoded) << '\n';
    return 0;
}
